/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>  /* for getuid/seteuid */
#endif
#include "nsDebug.h"
#include "plstr.h"
#include "ifdetect.h"
#include "dhcpinform.h"

#define UNICAST_WAIT_TIME 1
#define TOTAL_WAIT_TIME 5

/* some DHCP-related constants */

/* offsets of things */
static const PRUint32 kOpOffset = 0;
static const PRUint32 kHTypeOffset = 1;
static const PRUint32 kHLenOffset = 2;
static const PRUint32 kHOpsOffset = 3;
static const PRUint32 kXidOffset = 4;
static const PRUint32 kSecsOffset = 8;
static const PRUint32 kFlagsOffset = 10;
static const PRUint32 kCIAddrOffset = 12;
static const PRUint32 kYIAddrOffset = 16;
static const PRUint32 kSIAddrOffset = 20;
static const PRUint32 kGIAddrOffset = 24;
static const PRUint32 kCHAddrOffset = 28;
static const PRUint32 kSnameOffset = 44;
static const PRUint32 kSnameLen = 64;
static const PRUint32 kFileOffset = 108; /* SnameOffset + SnameLen */
static const PRUint32 kFileLen = 128;
static const PRUint32 kCookieOffset = 236; /* FileOffset + FileLen */
static const PRUint32 kOptionsOffset = 240; /* CookieOffset+sizeof(OptionsCookie) */
static const PRUint32 kMinDHCPLen = 240; /* OptionsOffset */

/* DHCP operations */
static const PRUint8 kOpRequest = 1;
static const PRUint8 kOpReply = 2;

/* the magic cookie that defines the start of options */
static const PRUint8 kOptionsCookie[4] = { 99, 130, 83, 99 };

/* DHCP options that need symbolic names */
static const PRUint8 kOptionPad = 0;
static const PRUint8 kOptionOverload = 52;
static const PRUint8 kOptionMessageType = 53;
static const PRUint8 kOptionParameterRequest = 55;
static const PRUint8 kOptionMaxDHCPMessage = 57;
static const PRUint8 kOptionEnd = 255;

/* meanings of the overload option */
static const PRUint32 kOverloadNone = 0x0; /* not specified, useful default value */
/* The following two aren't specified as flags, but conveniently, if both
 * fields are overloaded, the value is 3, so I'll use them as flags.
 */
static const PRUint32 kOverloadFile = 0x1;
static const PRUint32 kOverloadSname = 0x2;

/* meanings of the message type option */
static const PRUint8 kMessageTypeDiscover = 1;
static const PRUint8 kMessageTypeOffer = 2;
static const PRUint8 kMessageTypeRequest = 3;
static const PRUint8 kMessageTypeDecline = 4;
static const PRUint8 kMessageTypeACK = 5;
static const PRUint8 kMessageTypeNAK = 6;
static const PRUint8 kMessageTypeRelease = 7;
static const PRUint8 kMessageTypeInform = 8;

/* port numbers */
static const PRUint16 kPortServer = 67;
static const PRUint16 kPortClient = 68;

/* Looks at the option pointed to by option, and interprets it as a single
 * octet meaning message type.  Saves the type in *messageType (if messageType
 * is not nsnull).
 * Assumes option is not nsnull, and len is 1.
 */
static void parseMessageType(PRUint32 code, const PRUint8 *option,
 PRUint32 len, PRUint8 *messageType)
{
    NS_ASSERTION(option, "missing option data");
    if (!option)
        return;
    if (len > 0) {
        if (messageType)
            *messageType = *option;
    }
}

/* Interprets the option data at *option as the overload option, and saves it
 * in *overload (if overload is not nsnull).
 * Assumes option is not nsnull, and len is 1.
 */
static void parseOverloadOption(PRUint32 code, const PRUint8 *option,
 PRUint32 len, PRUint8 *overload)
{
    NS_ASSERTION(option, "missing option data");
    if (!option)
        return;
    if (len > 0) {
        if (overload)
            *overload = *option;
    }
}

/* Parses the options starting at buf, first parsing built-in options, and
 * calling callback (if callback is not nsnull) for unknown options.  Fills in
 * messageType with the value of the MessageType option, and overload with the
 * value of the Overload option, if each is not nsnull.
 * Assumes buf is not nsnull.  Continues parsing until len bytes have been read,
 * or the end option has been found.
 */
static void parseOptions(const PRUint8 *buf, PRUint32 len,
 OptionCallback callback, PRUint8 *messageType, PRUint8 *overload)
{
    PRUint32 offset;
    PRBool end = PR_FALSE;

    NS_ASSERTION(buf, "missing option data");
    if (!buf)
        return;

    for (offset = 0; offset < len && !end; ) {
        PRUint8 option = buf[offset], optionLen;

        if (offset + 1 < len)
            optionLen = PR_MIN(len - offset - 2, *(buf + offset + 1));
        else
            optionLen = 0;
        switch (option)
        {
            case kOptionPad:
                offset++;
                break;
            case kOptionEnd:
                end = PR_TRUE;
                break;
            case kOptionMessageType:
                parseMessageType(option, buf + offset + 2, optionLen,
                 messageType);
                offset += 2 + optionLen;
                break;
            case kOptionOverload:
                parseOverloadOption(option, buf + offset + 2, optionLen,
                 overload);
                offset += 2 + optionLen;
                break;
            default:
                if (option < MAX_DHCP_OPTION && callback)
                    callback(option, buf + offset + 2, optionLen);
                offset += 2 + optionLen;
        }
    }
}

/* Parses the DHCP packet using the parsers contained in context, if the
 * packet's XID matches XID.
 * parser is the callback function that will be called with every encountered
 * DHCP option.  If it's NULL, you won't hear about any options--kinda
 * pointless, but allowed.
 * Returns whether the parsed packet is an ACK whose XID matches XID (and
 * therefore is the response we're looking for).
 * Assumes context and pkt are not NULL.
 */
static PRBool parseDHCP(PRUint32 XID, const PRUint8 *pkt, PRUint32 len,
 OptionCallback parser)
{
    PRBool done = PR_FALSE;
    PRUint32 xid;

    NS_ENSURE_TRUE(pkt, PR_FALSE);
    NS_ENSURE_TRUE(len >= kMinDHCPLen, PR_FALSE);

    memcpy(&xid, pkt + kXidOffset, sizeof(xid));
    /* do a host-order comparison, as caller's providing us with xid to match */
    xid = PR_ntohl(xid);
    if (xid == XID) {
        PRUint8 messageType = 0, overload = kOverloadNone;

        /* do a once-over to get MessageType and Overload without calling parser
         * first, then go over again with parser.
         */
        parseOptions(pkt + kOptionsOffset, len - kOptionsOffset,
         nsnull, &messageType, &overload);
        if (messageType == kMessageTypeACK) {
            parseOptions(pkt + kOptionsOffset, len - kOptionsOffset,
             parser, nsnull, nsnull);
            if (overload & kOverloadFile)
                parseOptions(pkt + kFileOffset, kFileLen, parser, nsnull,
                 nsnull);
            if (overload & kOverloadSname)
                parseOptions(pkt + kSnameOffset, kSnameLen, parser, nsnull,
                 nsnull);
            done = PR_TRUE;
        }
    }
    return done;
}

/* Adds an option to the DHCP packet at address buf.  bufLen indicates how
 * many bytes may be stored at buf.  currentLen is the current number of
 * bytes stored at buf.  option is the option code, optionLen is the number
 * of bytes from optionData to copy.
 * Assumes buf is not nsnull, bufLen is at the very least MinDHCPLen bytes.
 * If option is 0 or 255, only the option code is stored, and currentLen + 1
 * must be less than bufLen.
 * If option is any other value, the option code and optionLen are stored,
 * and currentLen + optionLen + 2 must be less than bufLen.
 * If optionLen is greater than 0, assumes that optionData is not nsnull.
 * Returns 0 on success (after which the caller should take care to update
 * currentLen to currentLen + optionLen + 2), -1 on failure.
 */
static int addOption(PRUint8 *buf, PRUint32 bufLen, PRUint32 currentLen,
 PRUint32 option, PRUint32 optionLen, const PRUint8 *optionData)
{
    NS_ENSURE_TRUE(buf, -1);
    NS_ENSURE_TRUE(bufLen > kMinDHCPLen, -1);
    NS_ENSURE_TRUE(currentLen + 1 <= bufLen, -1);

    buf[currentLen] = option;
    if (option != 0 && option != 255) {
        NS_ENSURE_TRUE(currentLen + optionLen + 2 <= bufLen, -1);
        buf[currentLen + 1] = optionLen;
        if (optionLen > 0) {
            NS_ENSURE_TRUE(optionData, -1);
            memcpy(buf + currentLen + 2, optionData, optionLen);
        }
    }
    return 0;
}

/* Creates the DHCPINFORM request for us at buffer buf with length *bufLen.
 * Uses XID as the transaction ID, and info's ip and hwAddr as the local IP
 * address and hardware address, respectively.  maxResponseLen is the maximum
 * size of DHCP response you (the caller) are prepared to receive, in host
 * order.  Requests all options.
 * Assumes buf and bufLen are not NULL, and *bufLen is at the very least
 * MinDHCPLen long.
 * On success, *bufLen will be the length of the request in buf.
 * Returns 0 on success and -1 on failure.
 */
static int createRequest(PRUint8 *buf, PRUint16 *bufLen, PRUint32 XID,
 const struct InterfaceInfo *info, PRUint32 maxResponseLen)
{
    int ret;
    PRUint16 currentLen = kMinDHCPLen, maxLength;
    PRUint8 op, optionsToRequest[MAX_DHCP_OPTION];
    PRUintn i;

    NS_ENSURE_TRUE(buf, -1);
    NS_ENSURE_TRUE(bufLen, -1);
    NS_ENSURE_TRUE(*bufLen > kMinDHCPLen, -1);
    NS_ENSURE_TRUE(info, -1);
    NS_ENSURE_TRUE(info->hwAddrLen <= kSnameOffset - kCHAddrOffset, -1);
    NS_ENSURE_TRUE(maxResponseLen <= PR_UINT16_MAX, -1);

    buf[kOpOffset]    = kOpRequest;
    buf[kHTypeOffset] = (PRUint8)info->hwType;
    buf[kHLenOffset]  = info->hwAddrLen;
    buf[kHOpsOffset]  = 0;
    XID = PR_htonl(XID);
    memcpy(buf + kXidOffset, &XID, sizeof(XID));
    memset(buf + kSecsOffset, 0, sizeof(PRUint16)); /* secs since start */
    memset(buf + kFlagsOffset, 0, sizeof(PRUint16)); /* we accept unicast */
    memcpy(buf + kCIAddrOffset, &info->ip, sizeof(info->ip));
    memset(buf + kYIAddrOffset, 0, sizeof(PRUint32)); /* yiaddr */
    memset(buf + kSIAddrOffset, 0, sizeof(PRUint32)); /* siaddr */
    memset(buf + kGIAddrOffset, 0, sizeof(PRUint32)); /* giaddr */
    memcpy(buf + kCHAddrOffset, info->hwAddr, info->hwAddrLen);
    memset(buf + kSnameOffset, 0, kSnameLen);
    memset(buf + kFileOffset, 0, kFileLen);
    memcpy(buf + kCookieOffset, kOptionsCookie, sizeof(kOptionsCookie));

    op = kMessageTypeInform;
    ret = addOption(buf, *bufLen, currentLen, kOptionMessageType, sizeof(op),
     &op);
    NS_ENSURE_TRUE(ret == 0, -1);
    currentLen += sizeof(op) + 2;
    maxLength = PR_htons((PRUint16)maxResponseLen);
    ret = addOption(buf, *bufLen, currentLen, kOptionMaxDHCPMessage,
     sizeof(maxLength), (const PRUint8 *)&maxLength);
    NS_ENSURE_TRUE(ret == 0, -1);
    currentLen += sizeof(maxLength) + 2;

    for (i = 0; i < MAX_DHCP_OPTION; i++)
      optionsToRequest[i] = i;
    ret = addOption(buf, *bufLen, currentLen, kOptionParameterRequest,
     sizeof(optionsToRequest), optionsToRequest);
    NS_ENSURE_TRUE(ret == 0, -1);
    currentLen += sizeof(optionsToRequest) + 2;

    ret = addOption(buf, *bufLen, currentLen, kOptionEnd, 0, nsnull);
    NS_ENSURE_TRUE(ret == 0, -1);
    currentLen += 0 + 1;

    *bufLen = currentLen;

    return ret;
}

/* Prepares a socket to send a DHCP request with local IP address localaddr,
 * and ready for broadcast if broadcast is set.
 * Assumes localaddr is in network order.
 * Returns the socket descriptor if successful, nsnull on error.
 */
static PRFileDesc *prepareSocket(PRUint32 localaddr, PRBool broadcast)
{
    PRFileDesc *ret = PR_NewUDPSocket();

    if (ret) {
        PRSocketOptionData opt;
        PRStatus rv;

        opt.option = PR_SockOpt_Reuseaddr;
        opt.value.reuse_addr = PR_TRUE;

        /* in case a DHCP client is running, reuse the socket */
        rv = PR_SetSocketOption(ret, &opt);
        if (broadcast) {
            opt.option = PR_SockOpt_Broadcast;
            opt.value.broadcast = PR_TRUE;
            rv = PR_SetSocketOption(ret, &opt);
            if (NS_FAILED(rv)) {
                PR_Close(ret);
                ret = nsnull;
            }
        }
        if (NS_SUCCEEDED(rv)) {
            PRNetAddr addr;

            memset(&addr, 0, sizeof(addr));
            addr.inet.family = PR_AF_INET;
            addr.inet.port = PR_htons(kPortClient);
            addr.inet.ip = localaddr;

            raisePermissions();
            rv = PR_Bind(ret, &addr);
            dropPermissions();
            if (NS_FAILED(rv)) {
                PR_Close(ret);
                ret = nsnull;
            }
        }
    }
    return ret;
}

/* This function:
 * - prepares a socket on which to send a DHCPINFORM request (bound to
 *   info's ip) to addr
 * - calls createRequest to create the message, with XID as the DHCP xid,
 *   and sends it
 * recvBufLen is the maximum number of bytes you're able to receive.
 * Assumes info is not nsnull, and that recvBufLen <= PR_UINT16_MAX.
 * Returns the socket on which the request was sent on success, nsnull on
 * error.
 */
static PRFileDesc *sendDHCPInform(const struct InterfaceInfo *info,
 PRUint32 XID, PRUint32 server, size_t recvBufLen)
{
    PRFileDesc *sock;

    NS_ENSURE_TRUE(info, nsnull);
    NS_ENSURE_TRUE(recvBufLen <= PR_UINT16_MAX, nsnull);

    sock = prepareSocket(info->ip, server == info->bcastAddr);
    if (sock) {
        PRUint8 msg[2048];
        PRUint16 msgLen = sizeof(msg);

        if (createRequest(msg, &msgLen, XID, info, recvBufLen) == 0) {
            PRNetAddr addr;
            PRInt32 sendtoReturn;

            addr.inet.family = PR_AF_INET;
            addr.inet.port = PR_htons(kPortServer);
            addr.inet.ip = server;
            sendtoReturn = PR_SendTo(sock, msg, msgLen, 0, &addr,
             PR_INTERVAL_NO_WAIT);
            if (sendtoReturn < msgLen) {
                PR_Close(sock);
                sock = nsnull;
            }
        }
        else {
            perror("createRequest");
            PR_Close(sock);
            sock = nsnull;
        }
    }
    return sock;
}

/* Sends requests.  Destinations depend on retryBroadcasts:
 * - if False, assumes a nonzero sock member of an InterfaceInfo implies the
 *   interface has been selected for use.
 * - if True, assumes a nonzero sock member of an InterfaceInfo implies
 *   the InterfaceInfo has been used to send a request, and the (sock - 1) is
 *   an index into toPoll (and must therefore be less than toPollLen as well).
 *   If the InterfaceInfo's serverAddr member is not INADDR_ANY, then a unicast
 *   has been sent, and the existing socket is closed, and a new request is
 *   sent to the bcastAddr address instead.
 * toPoll is an in/out parameter.
 * numToPoll is also in/out.  Assumes its current value is the number of
 * entries in toPoll that are valid.  On return, its value is the new number,
 * if different.
 * XID is the DHCP transaction ID to use.
 * maxReponseLen is the maximum length of response you're able to receive.
 * Assumes toPoll and numToPoll are not NULL, and toPollLen is positive.
 * Returns 0 on success, -1 on failure.
 * Warning:  the use of sock is a hack, and rather easy to f*** up.
 * Suggestions for improvement welcome.
 */
static int sendRequests(struct InterfaceInfo *interfaces, PRPollDesc *toPoll,
 PRUintn toPollLen, PRUintn *numToPoll, PRBool *useUnicast,
 PRBool retryBroadcasts, PRUint32 XID, PRUintn maxResponseLen)
{
    int ret;
    struct InterfaceInfo *ptr;

    NS_ENSURE_TRUE(toPoll, -1);
    NS_ENSURE_TRUE(toPollLen > 0, -1);
    NS_ENSURE_TRUE(numToPoll, -1);

    ret = 0;
    for (ptr = interfaces; ptr && ret == 0; ptr = ptr->next) {
        if (ptr->sock != 0) {
            PRBool sendOnThisInterface = PR_FALSE;
            PRUint32 server = PR_INADDR_ANY;

            if (retryBroadcasts) {
                if (ptr->sock > 0 && ptr->serverAddr != PR_INADDR_ANY) {
                    PRUintn toPollNdx = ptr->sock - 1;

                    if (toPollNdx < toPollLen && toPoll[toPollNdx].fd) {
                        PR_Close(toPoll[toPollNdx].fd);
                        toPoll[toPollNdx].fd = nsnull;
                        ptr->sock = 0;
                        sendOnThisInterface = PR_TRUE;
                        server = ptr->bcastAddr;
                        *numToPoll = *numToPoll - 1;
                    }
                }
            }
            else {
                sendOnThisInterface = PR_TRUE;
                if (ptr->serverAddr != PR_INADDR_ANY) {
                    if (useUnicast)
                        *useUnicast = PR_TRUE;
                    server = ptr->serverAddr;
                }
                else
                    server = ptr->bcastAddr;
            }
            if (sendOnThisInterface) {
                toPoll[*numToPoll].fd = sendDHCPInform(ptr, XID, server,
                 maxResponseLen);
                if (toPoll[*numToPoll].fd) {
                    ptr->sock = *numToPoll + 1;
                    toPoll[*numToPoll].in_flags = PR_POLL_READ;
                    *numToPoll = *numToPoll + 1;
                }
                else
                    ret = -1;
            }
        }
    }
    return ret;
}

/* Waits on toPoll, which is an array ot numToPoll poll descriptors, until the
 * absolute time specified in waitUntil passes, or a packet whose transaction ID
 * matches XID is received.
 * Receives into buf (with length bufLen).
 * Assumes toPoll, buf, and done are not NULL, and numToPoll is positive.
 * Returns 0 on timeout/success and -1 on failure.
 * On return, *done is set to True if XID is matched.
 * Apologies on the interface:  buf should be able to be local, except bufLen
 * was needed to call sendRequests, so I let you declare it statically and
 * pass it and its length here.
 */
static PRInt32 doWait(PRPollDesc *toPoll, PRUintn numToPoll,
 PRIntervalTime waitUntil, PRUint32 XID, PRUint8 *buf, PRUint32 bufLen,
 PRBool *done, OptionCallback cb)
{
    PRInt32 ret;
    PRIntervalTime timeToWait = waitUntil - PR_IntervalNow();

    NS_ENSURE_TRUE(toPoll, -1);
    NS_ENSURE_TRUE(numToPoll > 0, -1);
    NS_ENSURE_TRUE(buf, -1);
    NS_ENSURE_TRUE(done, -1);

    if (timeToWait > 0) {
        do {
            ret = PR_Poll(toPoll, numToPoll, timeToWait);
            if (ret > 0) {
                PRUintn i;

                for (i = 0; i < numToPoll; i++) {
                    if (toPoll[i].out_flags & PR_POLL_READ) {
                        PRNetAddr server;
                        PRInt32 bytesReceived = PR_RecvFrom(toPoll[i].fd,
                         buf, bufLen, 0, &server, PR_INTERVAL_NO_WAIT);

                        if (bytesReceived > 0)
                            *done = parseDHCP(XID, buf, bytesReceived, cb);
                    }
                }
            }
        } while (ret > 0 && !*done);
        if (done)
            ret = 0;
    }
    else
        ret = 0;
    return ret;
}

/* Gets a value for an XID using PR_GetRandomNoise.  This doesn't make any
 * guarantee of randomness, so if it doesn't modify the value, we have whatever
 * value ret is given, and that's okay too.  There's no cryptographic
 * requirement for this thing nor is there a session to hijack.
 */
static PRUint32 getXID(void)
{
    PRUint32 ret;

    PR_GetRandomNoise(&ret, sizeof(ret));
    return ret;
}

int doDHCPInform(struct InterfaceInfo *interfaces, OptionCallback cb)
{
    int ret;
    PRUint8 buf[2048];
    PRBool useUnicast;
    PRPollDesc *toPoll;
    PRUintn numToPoll, numDestinations;
    PRUint32 XID;
    struct InterfaceInfo *ptr;

    numDestinations = 0;
    for (ptr = interfaces; ptr; ptr = ptr->next)
        if (ptr->sock != 0)
            numDestinations++;
    toPoll = (PRPollDesc *)PR_Calloc(1, sizeof(PRPollDesc) * numDestinations);
    NS_ENSURE_TRUE(toPoll, -1);

    numToPoll = 0;
    useUnicast = PR_FALSE;
    XID = getXID();
    ret = sendRequests(interfaces, toPoll, numDestinations, &numToPoll,
     &useUnicast, PR_FALSE, XID, sizeof(buf));
    if (ret == 0 && numToPoll > 0) {
        PRIntervalTime waitUntil = PR_IntervalNow() + PR_SecondsToInterval(
         useUnicast ? UNICAST_WAIT_TIME : TOTAL_WAIT_TIME);
        PRBool done = PR_FALSE;

        ret = doWait(toPoll, numToPoll, waitUntil, XID, buf, sizeof(buf),
         &done, cb);
        if (ret == 0 && !done && useUnicast) {
            /* timeout, retry all unicasts as broadcasts */
            ret = sendRequests(interfaces, toPoll, numDestinations, &numToPoll,
             nsnull, PR_TRUE, XID, sizeof(buf));
            if (ret == 0) {
                waitUntil = PR_IntervalNow() +
                 PR_SecondsToInterval(TOTAL_WAIT_TIME - UNICAST_WAIT_TIME);
                ret = doWait(toPoll, numToPoll, waitUntil, XID, buf,
                 sizeof(buf), &done, cb);
            }
        }
    }
    PR_Free(toPoll);
    return ret;
}

#ifdef XP_UNIX
void dropPermissions(void)
{
    seteuid(getuid());
}

void raisePermissions(void)
{
    seteuid(0);
}
#endif
