/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
/* This program implements the DHCPINFORM portion of DHCP (see RFC2131) to get
 * DHCP options.  It prints the data from the DHCP server in ASCII form to
 * stdout.
 * The program's intended to play nice, therefore:
 * - I'm careful about length checking to avoid buffer overruns
 * - since I expect I need to be root under UNIX to bind to the DHCP client
 *   port, I use seteuid() to hold root's euid as little as possible
 * - I allow you to specify which interface to talk through and which DHCP
 *   server you want to try, so you can be nice to your neighbors and avoid
 *   unnecessary broadcasts
 * - I've separated the things I know are platform-specific into functions or
 *   macros and marked them with <platform>-specific
 * See TODO's for stuff I think isn't quite there yet.
 */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nsDebug.h" /* for NS_ASSERTION/NS_ENSURE_TRUE */
#include "plstr.h"   /* for PL_strcasecmp */
#include "prio.h"
#include "ifdetect.h"
#include "dhcpinform.h"

/* non-const globals */
static PRBool gVerbose = 0;
static PRBool gQuiet = 0;
static PRBool gHelp = 0;
static struct InterfaceInfo *gInterfaces = nsnull;
static PRUintn gNumDestinations = 0;

static void parseOptionCallback(PRUint8 code, const PRUint8 *option,
 PRUint16 len)
{
    PRUint16 i;

    NS_ASSERTION(option, "option data is null");
    if (!option)
        return;
    printf("%d: ", (PRUintn)code);
    for (i = 0; i < len; i++)
      printf("%02x ", option[i]);
    putchar('\n');
}

/* Prints the usage of this program; arg is assumed to be a NULL-terminated
 * string representing the program name (argv[0] will do just fine).
 */
static void printUsage(const char *arg)
{
    NS_ASSERTION(arg, "missing argument");
    if (!arg)
        return;
    printf(
     "Usage: %s [-d dest]* [-q | -v] [-h]\n",
     arg);
    printf(
     "  -d: destination to send to, in the form [server]@<interface>, where:\n");
    printf(
     "      - server is IP address of the DHCP server\n");
    printf(
     "      - interface is the interface with which to send, e.g. eth0\n");
    printf(
     "      If the server is specified, sends a unicast packet to server\n");
    printf(
     "      using the specified interface.  If not specified, sends a\n");
    printf(
     "      broadcast to the interface's broadcast address.  If no\n");
    printf(
     "      destinations are specified, will broadcast to every interface\n");
    printf(
     "      that supports it, and return options from the first server that\n");
    printf(
     "      responds.\n\n");
    printf(
     "      Example 1:  '-d @eth0' sends a broadcast to eth0\n");
    printf(
     "      Example 2:  '-d 192.168.1.1@eth0' sends a unicast to 192.168.1.1\n");
    printf(
     "                  on interface eth0\n\n");
    printf(
     "  -q: quiet mode (don't print option codes/names, just values)\n");
    printf(
     "  -v: verbose mode\n\n");
    printf(
     "  -h:             print usage (this output)\n");
    printf(
     "\nIf you see an error like, \"bind: Permission denied\", you need to\n");
    printf(
     "run as root, or make the program suid root.  bind(2) needs to be root\n");
    printf(
     "to bind to the DHCP client port.\n");
}

/* Checks whether arg is a valid interface in the list gInterfaces.  Returns
 * the associated struct InterfaceInfo * if so, and nsnull if not.
 * Assumes name is not nsnull.
 */
static struct InterfaceInfo *findInterface(const char *name)
{
    struct InterfaceInfo *ret = nsnull, *ptr;

    NS_ENSURE_TRUE(name, nsnull);

    for (ptr = gInterfaces; ptr && !ret; ptr = ptr->next) {
        if (!PL_strcasecmp(ptr->name, name))
            ret = ptr;
    }
    return ret;
}

/* Parses the argument arg as a destination of the form [server]@<interface>.
 * If it parses incorrectly, prints to stderr.  Otherwise stores the parsed
 * destination information in the related struct Interface in gInterfaces, and
 * marks it in use by setting its sock member to 1.  Also increments
 * gNumDestinations for every valid interface found.
 * Assumes arg is not nsnull.
 */
static void parseDestination(char *arg)
{
    NS_ASSERTION(arg, "missing argument");
    if (arg) {
        char *sep = PL_strchr(arg, '@');
    
        if (!sep)
            fprintf(stderr, "-d: bad syntax, expect [server]@<interface>\n");
        else
        {
            struct InterfaceInfo *iface = findInterface(sep + 1);

            if (iface) {
                if (sep != arg) {
                    PRNetAddr addr;

                    *sep = '\0';
                    if (NS_FAILED(PR_StringToNetAddr(arg, &addr)))
                        fprintf(stderr, "%s: invalid server IP address\n", arg);
                    else {
                        iface->serverAddr = addr.inet.ip;
                        iface->sock = 1;
                        gNumDestinations++;
                    }
                }
                else {
                    iface->sock = 1;
                    gNumDestinations++;
                }
            }
            else
                fprintf(stderr, "%s: invalid interface\n", sep + 1);
        }
    }
}

/* Parses the command line (argc and argv) for arguments.  As side effects,
 * parserTable, gInterfaces, and gNumInterfaces may be modified.
 * See parseOptionCode, parseDestination, and requestAllOptions for more
 * details.
 */
static void parseCmdLine(int argc, char *argv[])
{
    int i;

    for (i = 1; i < argc && argv[i]; ) {
        if (*argv[i] == '-') {
            switch (*(argv[i] + 1))
            {
                case 'h':
                    gHelp = PR_TRUE;
                    break;
                case 'd':
                    if (i + 1 < argc)
                    {
                        parseDestination(argv[i + 1]);
                        i++;
                    }
                    else
                        fprintf(stderr, "-d: required argument missing\n");
                    break;
                case 'q':
                    gQuiet = PR_TRUE;
                    if (gVerbose)
                        fprintf(stderr, "Can't be both verbose and quiet\n");
                    break;
                case 'v':
                    gVerbose = PR_TRUE;
                    if (gQuiet)
                        fprintf(stderr, "Can't be both verbose and quiet\n");
                    break;
                default:
                    fprintf(stderr, "Unknown option -%c\n", *(argv[i] + 1));
            }
        }
        i++;
    }
}

/* Marks all interfaces in gInterfaces in use by setting their sock member
 * to 1.  Used to simplify mainLoop, which will only have to iterate over
 * gInterfaces and not special-case the no-destinations-specified case.
 */
void enableAllDestinations(void)
{
    struct InterfaceInfo *ptr;

    for (ptr = gInterfaces; ptr; ptr = ptr->next) {
        ptr->sock = 1;
        gNumDestinations++;
    }
}

int main(int argc, char *argv[])
{
    int ret;

    dropPermissions();
    gInterfaces = findDHCPInterfaces();

    parseCmdLine(argc, argv);
    if (gHelp) {
        printUsage(argv[0]);
        ret = 0;
    }
    else {
        if (gNumDestinations == 0)
            enableAllDestinations();
        ret = doDHCPInform(gInterfaces, parseOptionCallback);
    }

    freeInterfaceList(gInterfaces);
    gInterfaces = nsnull;
    return ret;
}
