/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <stdio.h>
#include <windef.h>
#include <winbase.h>
#include <iphlpapi.h>
#include "nsDebug.h" /* for NS_ASSERTION etc. */
#include "plstr.h"   /* for PL_strdup */
#include "prio.h"    /* for PR_INADDR_ANY */
#include "ifdetect.h"

typedef DWORD (WINAPI *GetAdaptersInfoFunc)(PIP_ADAPTER_INFO, PULONG);
typedef DWORD (WINAPI *GetInterfaceInfoFunc)(PIP_INTERFACE_INFO, PULONG);

static HMODULE gIPHelper = nsnull;
static GetAdaptersInfoFunc gGetAdaptersInfo = nsnull;
static GetInterfaceInfoFunc gGetInterfaceInfo = nsnull;

static void initLibrary(void)
{
    if (!gIPHelper)
        gIPHelper = LoadLibraryA("iphlpapi.dll");
    if (gIPHelper) {
        gGetAdaptersInfo = (GetAdaptersInfoFunc)GetProcAddress(gIPHelper,
         "GetAdaptersInfo");
        gGetInterfaceInfo = (GetInterfaceInfoFunc)GetProcAddress(gIPHelper,
         "GetInterfaceInfo");
    }
}

static void freeLibrary(void)
{
    if (gIPHelper) {
        FreeLibrary(gIPHelper);
        gIPHelper = nsnull;
        gGetAdaptersInfo = nsnull;
        gGetInterfaceInfo = nsnull;
    }
}

/* Returns a NULL-terminated string representing the adapter's name.  Assumes
 * adapter and map are not nsnull.
 * A note on implemenation:  the obvious approach uses adapter's AdapterName
 * member, but this is only filled in on WinNTish systems, ergo the use of the
 * map on Win9xish systems.
 */
static char *getAdapterName(PIP_ADAPTER_INFO adapter, PIP_INTERFACE_INFO map)
{
    char *ret = nsnull;
    OSVERSIONINFO info;

    NS_ENSURE_TRUE(adapter, nsnull);
    NS_ENSURE_TRUE(map, nsnull);

    memset(&info, 0, sizeof(info));
    info.dwOSVersionInfoSize = sizeof(info);
    GetVersionEx(&info);
    if (info.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) {
        LONG i;

        for (i = 0; i < map->NumAdapters && !ret; i++) {
            if (map->Adapter[i].Index == adapter->Index)
                ret = PL_strdup((char *)map->Adapter[i].Name);
        }
    }
    else
        ret = PL_strdup(adapter->AdapterName);
    return ret;
}

struct InterfaceInfo *findDHCPInterfaces(void)
{
    struct InterfaceInfo *interfaces = nsnull;

    initLibrary();
    if (gGetAdaptersInfo && gGetInterfaceInfo) {
        ULONG adaptersLen = 0;
        PIP_ADAPTER_INFO adapters = nsnull;

        gGetAdaptersInfo(nsnull, &adaptersLen);
        adapters = (PIP_ADAPTER_INFO)PR_Malloc(adaptersLen);
        if (gGetAdaptersInfo(adapters, &adaptersLen) == ERROR_SUCCESS) {
            PIP_INTERFACE_INFO map = nsnull;
            ULONG mapLen = 0;

            gGetInterfaceInfo(nsnull, &mapLen);
            map = (PIP_INTERFACE_INFO)PR_Malloc(mapLen);
            if (gGetInterfaceInfo(map, &mapLen) == ERROR_SUCCESS) {
                PIP_ADAPTER_INFO ptr = nsnull;

                for (ptr = adapters; ptr; ptr = ptr->Next) {
                    PRNetAddr addr;
                    PRStatus rv;

                    rv = PR_StringToNetAddr(ptr->IpAddressList.IpAddress.String,
                     &addr);
                    /* if there's no IP, we can't bind(), so don't bother */
                    if (ptr->DhcpEnabled && NS_SUCCEEDED(rv) && addr.inet.ip !=
                     PR_INADDR_ANY) {
                        struct InterfaceInfo *newIF = (struct InterfaceInfo *)
                         PR_Calloc(1, sizeof(struct InterfaceInfo));
                        PRUint32 mask;

                        newIF->name = getAdapterName(ptr, map);
                        switch (ptr->Type)
                        {
                            case IF_TYPE_ETHERNET_CSMACD:
                                newIF->hwType = InterfaceHWTypeEthernet;
                                break;
                            default:
                                newIF->hwType = InterfaceHWTypeUnknown;
                        }
                        newIF->hwAddrLen = ptr->AddressLength;
                        memcpy(newIF->hwAddr, ptr->Address,
                         PR_MAX(ptr->AddressLength, MAX_DHCP_HWADDR));
                        newIF->ip = addr.inet.ip;
                        rv = PR_StringToNetAddr(ptr->IpAddressList.IpMask.
                         String, &addr);
                        if (NS_SUCCEEDED(rv))
                            mask = addr.inet.ip;
                        else
                            mask = PR_INADDR_ANY;
                        newIF->bcastAddr = (newIF->ip & mask) | ~mask;
                        rv = PR_StringToNetAddr(ptr->DhcpServer.IpAddress.
                         String, &addr);
                        if (NS_SUCCEEDED(rv))
                            newIF->serverAddr = addr.inet.ip;
                        else
                            newIF->serverAddr = PR_INADDR_ANY;
                        newIF->next = interfaces;
                        interfaces = newIF;
                    }
                }
            }
        }
        PR_Free(adapters);
    }
    freeLibrary();
    return interfaces;
}
