/* -*- Mode: IDL; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "nsAString.h"
#include "nsDHCPOptionService.h"

NS_IMPL_THREADSAFE_ISUPPORTS1(nsDHCPOptionService, nsIDHCPOptionService)

// XXX bluh, how many times I gotta write silly classes like this?  Need it to
// store in the nsHashtable, fuggitall..
class DHCPOptionEntry {
public:
    DHCPOptionEntry(const PRUint8 *aData, PRUint16 aLen)
     : mLen(0), mDataSize(0), mData(0)
     { CopyData(aData, aLen); }
    DHCPOptionEntry(const DHCPOptionEntry &rhs)
     : mLen(0), mDataSize(0), mData(0)
     { CopyData(rhs.mData, rhs.mLen); }
    ~DHCPOptionEntry()
     { FreeData(); }
    DHCPOptionEntry &operator=(const DHCPOptionEntry &rhs)
     { if (&rhs != this) { FreeData(); CopyData(rhs.mData, rhs.mLen); } return *this; }
    void AddData(const PRUint8 *aData, PRUint16 aLen)
     { CopyData(aData, aLen, mLen); }
    PRUint16 GetLen(void) const { return mLen; }
    const PRUint8 *GetData(void) const { return mData; }

protected:
    void CopyData(const PRUint8 *aData, PRUint16 aLen, PRUint16 aOffset = 0);
    void FreeData()
     { if (mData) delete [] mData; mData = 0; mLen = mDataSize = 0; }
    void AdjustSize(PRUint16 aNewSize);

    PRUint16 mLen;
    PRUint16 mDataSize;
    PRUint8 *mData;
};

void
DHCPOptionEntry::CopyData(const PRUint8 *aData, PRUint16 aLen, PRUint16 aOffset)
{
    if (aLen + aOffset > mDataSize)
        AdjustSize(aLen + aOffset);
    memcpy(mData + aOffset, aData, aLen);
    if (aOffset + aLen > mLen)
        mLen = aOffset + aLen;
}

void
DHCPOptionEntry::AdjustSize(PRUint16 aNewSize)
{
    if (aNewSize > mDataSize) {
        PRUint8 *data = new PRUint8[mDataSize = aNewSize];
        
        memcpy(data, mData, mLen);
        delete [] mData;
        mData = data;
    }
}

static PRBool
FreeDHCPOptionsMap(nsHashKey *aKey, void *aData, void *aClosure)
{
    DHCPOptionEntry *entry = (DHCPOptionEntry *)aData;
    delete entry;
    return PR_TRUE;
}

nsDHCPOptionService *nsDHCPOptionService::instance = nsnull;

nsDHCPOptionService::nsDHCPOptionService()
{
    instance = this;
}

nsDHCPOptionService::~nsDHCPOptionService()
{
    mOptionsMap.Reset(FreeDHCPOptionsMap, nsnull);
    instance = nsnull;
}

void
nsDHCPOptionService::OptionsCallback(PRUint8 aCode, const PRUint8 *aData,
 PRUint16 aLen)
{
    NS_ASSERTION(instance, "Options callback called with no class instance");
    nsPRUint32Key key(aCode);
    DHCPOptionEntry *entry = (DHCPOptionEntry *)instance->mOptionsMap.Get(&key);
    if (entry)
        entry->AddData(aData, aLen);
    else
        instance->mOptionsMap.Put(&key, new DHCPOptionEntry(aData, aLen));
}

nsresult
nsDHCPOptionService::InternalGetDHCPOption(PRUint8 aOption, PRUint16 *aLen,
 void **aData)
{
    if (!aLen) return NS_ERROR_NULL_POINTER;
    if (!aData) return NS_ERROR_NULL_POINTER;

    nsresult rv = NS_OK;
    nsPRUint32Key key(aOption);
    DHCPOptionEntry *entry = (DHCPOptionEntry *)mOptionsMap.Get(&key);
    if (entry) {
        *aData = new PRUint8[entry->GetLen()];
        if (!*aData)
            rv = NS_ERROR_OUT_OF_MEMORY;
        else {
            *aLen = entry->GetLen();
            memcpy(*aData, entry->GetData(), *aLen);
        }
    }
    else {
        *aLen = 0;
        *aData = 0;
    }
    return rv;
}

#if defined (XP_WIN)

#include "ifdetect.h"
#include "dhcpinform.h"

NS_IMETHODIMP
nsDHCPOptionService::GetDHCPOption(PRUint8 aOption, PRUint16 *aLen,
 void * *_retval)
{
    if (!aLen) return NS_ERROR_NULL_POINTER;
    if (!_retval) return NS_ERROR_NULL_POINTER;
    if (aOption > MAX_DHCP_OPTION) return NS_ERROR_INVALID_ARG;

    struct InterfaceInfo *interfaces = findDHCPInterfaces();
    if (interfaces) {
        // mark each interface "in use"
        for (struct InterfaceInfo *ptr = interfaces; ptr; ptr = ptr->next)
            ptr->sock = 1;

        // XXX I should be looking at the lease time of the DHCP interface;
        // maybe I should add this to InterfaceInfo?  Then I could cache these
        // values..
        if (doDHCPInform(interfaces, nsDHCPOptionService::OptionsCallback) == 0)
            InternalGetDHCPOption(aOption, aLen, _retval);
        // no point keeping these around..
        mOptionsMap.Reset(FreeDHCPOptionsMap);
        freeInterfaceList(interfaces);
    }
    return NS_OK;
}

#elif defined (XPUNIX)

#include "dhcpHelper.h"

NS_IMETHODIMP
nsDHCPOptionService::GetDHCPOption(PRUint8 aOption, PRUint16 *aLen,
 void * *_retval)
{
    if (!aLen) return NS_ERROR_NULL_POINTER;
    if (!_retval) return NS_ERROR_NULL_POINTER;
    if (aOption > MAX_DHCP_OPTION) return NS_ERROR_INVALID_ARG;

    nsresult rv;

    // If I knew the DHCP options' lifetime, I could cache these values.
    // However, the IP address lease time MUST NOT (rfc2131 table 3) be in a
    // DHCPINFORM.  Lame.  Probably could know by parsing DHCP lease files..
    rv = getDHCPOptionsFromHelper(nsDHCPOptionService::OptionsCallback);
    if (NS_SUCCEEDED(rv))
        rv = InternalGetDHCPOption(aOption, aLen, _retval);
    // no point keeping these around..
    mOptionsMap.Reset(FreeDHCPOptionsMap);
    return rv;
}
#else
NS_IMETHODIMP
nsDHCPOptionService::GetDHCPOption(PRUint8 aOption, PRUint16 *aLen,
 void * *_retval)
{
    // for folks porting:  shouldn't be that hard to get DHCP options, look at
    // the already implemented platforms for help.  The main thing is knowing
    // the MAC address of the interface you're sending on; the rest is just
    // gravy.  (If you know you're not running DHCP, for example, you can
    // short-circuit this whole thing, and if you know the DHCP server ID,
    // you can be a better netizen and unicast rather than broadcast.)
    return NS_ERROR_NOT_IMPLEMENTED;
}
#endif

NS_IMETHODIMP
nsDHCPOptionService::GetDHCPOptionAsCString(PRUint8 aOption,
 nsACString & _retval)
{
    PRUint16 len;
    void *data;
    nsresult rv;

    rv = GetDHCPOption(aOption, &len, &data);
    if (NS_SUCCEEDED(rv) && len > 0) {
        _retval.Assign((nsACString::char_type *)data, len);
        delete [] data;
    }
    return rv;
}
